/*
 * drivers/video/ep93xx/ep93xxfb_graphics.c
 *
 * Copyright 2009 Atmark Techno, Inc. All Rights Reserved.
 *
 * Based on the Cirrus Logic ep93xxfb driver.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/errno.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/arch/fb.h>
#include <asm/hardware.h>

static DECLARE_WAIT_QUEUE_HEAD(ep93xxfb_wait_in);

static irqreturn_t
ep93xxfb_irq_handler(int irq, void *dev_id)
{
	outl(0x00000000, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);
	wake_up(&ep93xxfb_wait_in);

	return IRQ_HANDLED;
}

int
ep93xxfb_request_irq(unsigned int irq)
{
	/* Add the Hardware accel irq */
	outl(0x00000000, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);
	return request_irq(irq, ep93xxfb_irq_handler, 0, "graphics", NULL);
}

void
ep93xxfb_free_irq(unsigned int irq)
{
	return free_irq(irq, NULL);
}

static void
ep93xxfb_wait(void)
{
	DECLARE_WAITQUEUE(wait, current);
	add_wait_queue(&ep93xxfb_wait_in, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);

	while (inl(EP93XX_GRAPHICS_ACCEL_BLOCKCTRL) & 0x00000001 )
		if (!in_atomic())
			schedule();

	remove_wait_queue(&ep93xxfb_wait_in, &wait);
	set_current_state(TASK_RUNNING);
}

static unsigned long
isqrt(unsigned long a)
{
	unsigned long rem = 0;
	unsigned long root = 0;
	int i;

	for (i = 0; i < 16; i++) {
		root <<= 1;
		rem = ((rem << 2) + (a >> 30));
		a <<= 2;
		root++;
		if (root <= rem) {
			rem -= root;
			root++;
		} else
			root--;
	}

	return root >> 1;
}

int
ep93xxfb_line(struct fb_info *info, struct ep93xx_line *line)
{
	unsigned long value = 0;
	long x, y, dx, dy, count, xinc, yinc, xval, yval, incr;

	if ((line->x1 > info->var.xres) ||
	    (line->x2 > info->var.xres) ||
	    (line->y1 > info->var.yres) ||
	    (line->y2 > info->var.yres))
		return -EFAULT;
	x = line->x1;
	y = line->y1;
	dx = line->x2 - line->x1;
	dy = line->y2 - line->y1;

	if ((dx < 0) && (dy < 0)) {
		x = line->x2;
		y = line->y2;
		dx *= -1;
		dy *= -1;
	} else if ((dx < 0) && (dy > 0)) {
		dx *= -1;
		value = 0x000000a0;
	} else if((dy < 0) && (dx > 0)) {
		dy *= -1;
		value = 0x00000140;
	}

	if (line->flags & LINE_PRECISE) {
		count = isqrt(((dy * dy) + (dx * dx)) * 4096);
		xinc = (4095 * 64 * dx) / count;
		yinc = (4095 * 64 * dy) / count;
		xval = 2048;
		yval = 2048;
		count = 0;
		while (dx || dy) {
			incr = 0;
			xval -= xinc;
			if (xval <= 0) {
				xval += 4096;
				dx--;
				incr = 1;
			}
			yval -= yinc;
			if (yval <= 0) {
				yval += 4096;
				dy--;
				incr = 1;
			}
			count += incr;
		}
	} else {
		if (dx == dy) {
			xinc = 4095;
			yinc = 4095;
			count = dx;
		} else if (dx < dy) {
			xinc = (dx * 4095) / dy;
			yinc = 4095;
			count = dy;
		} else {
			xinc = 4095;
			yinc = (dy * 4095) / dx;
			count = dx;
		}
	}

	outl(0x08000800, EP93XX_GRAPHICS_ACCEL_LINEINIT);
	if (line->flags & LINE_PATTERN)
		outl(line->pattern, EP93XX_GRAPHICS_ACCEL_LINEPATTRN);
	else
		outl(0x000fffff, EP93XX_GRAPHICS_ACCEL_LINEPATTRN);

	outl(epinfo.fb_phys + ((y * epinfo.xres * epinfo.bpp) / 8) +
	     ((x * epinfo.bpp) / 8), EP93XX_GRAPHICS_ACCEL_BLKDSTSTRT);

	outl(((x * epinfo.bpp) & 0x1F) | ((((x + dx - 1) * epinfo.bpp) & 0x1F ) << 16),
	     EP93XX_GRAPHICS_ACCEL_DESTPIXELSTRT);
	outl((epinfo.xres * epinfo.bpp) / 32,
	     EP93XX_GRAPHICS_ACCEL_DESTLINELENGTH);
	outl(line->fgcolor, EP93XX_GRAPHICS_ACCEL_BLOCKMASK);
	outl(line->bgcolor, EP93XX_GRAPHICS_ACCEL_BACKGROUND);
	outl((yinc << 16) | xinc, EP93XX_GRAPHICS_ACCEL_LINEINC);
	outl(count & 0xFFF, EP93XX_GRAPHICS_ACCEL_BLKDESTWIDTH);
	outl(0, EP93XX_GRAPHICS_ACCEL_BLKDESTHEIGHT);
	value |= (line->flags & LINE_BACKGROUND) ? 0x00004000 : 0;
	outl(epinfo.pixformat | value | 0x00000013,
	     EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);
	ep93xxfb_wait();

	return 0;
}

int
ep93xxfb_fillrect(struct fb_info *info, struct ep93xx_fill *fill)
{
	if ((fill->dx >= info->var.xres) ||
	    (fill->dy >= info->var.yres))
		return -EFAULT;

	if ((fill->dx + fill->width - 1) >= info->var.xres)
		fill->width = info->var.xres - fill->dx;
	if ((fill->dy + fill->height - 1) >= info->var.yres)
		fill->height = info->var.yres - fill->dy;

	outl(epinfo.fb_phys + (fill->dy * (info->var.xres * epinfo.bpp / 8)) +
	     ((fill->dx * epinfo.bpp / 8) & ~pixel_mask),
	     EP93XX_GRAPHICS_ACCEL_BLKDSTSTRT);
	outl(((fill->dx & pixel_mask) * epinfo.bpp) |
	     ((((fill->dx + fill->width - 1) & pixel_mask) *
	       epinfo.bpp) << 16), EP93XX_GRAPHICS_ACCEL_DESTPIXELSTRT);
	outl((info->var.xres * epinfo.bpp / 8) / 4,
	     EP93XX_GRAPHICS_ACCEL_DESTLINELENGTH);
	outl(((fill->dx + fill->width - 1) >> pixel_shift) -
	     (fill->dx >> pixel_shift), EP93XX_GRAPHICS_ACCEL_BLKDESTWIDTH);
	outl(fill->height - 1, EP93XX_GRAPHICS_ACCEL_BLKDESTHEIGHT);
	outl(fill->color, EP93XX_GRAPHICS_ACCEL_BLOCKMASK);
	outl(pixel_format | 0x0000000b, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);

	ep93xxfb_wait();

	return 0;
}

void
ep93xxfb_copyarea(struct fb_info *p, const struct fb_copyarea *area)
{
	unsigned long startsx, stopsx, startdx, stopdx, startsy, startdy;
	unsigned long blksrcwidth, blkdestwidth;
	unsigned long tmp;
	unsigned long val = 0;

	if (!area->width || !area->width ||
	    (area->sx >= p->var.xres) || (area->sy >= p->var.yres) ||
	    (area->dx >= p->var.xres) || (area->dy >= p->var.yres) ||
	    ((area->dx + area->width - 1) >= p->var.xres) ||
	    ((area->dy + area->height - 1) >= p->var.yres))
		return;

	if ((area->sx == area->dx) && (area->sy == area->dy))
		return;

	if ((area->dy == area->sy) && (area->dx > area->sx) &&
	    (area->dx < (area->sx + area->width - 1))) {
		startdx = area->dx + area->width - 1;
		stopdx = area->dx;
		startsx = area->sx + area->width - 1;
		stopsx = area->sx;
		val |= 0x000000a0;
	} else {
		startdx = area->dx;
		stopdx = area->dx + area->width - 1;
		startsx = area->sx;
		stopsx = area->sx + area->width - 1;
	}

	if (area->dy <= area->sy) {
		startdy = area->dy;
		startsy = area->sy;
	} else {
		startdy = area->dy + area->height -1;
		startsy = area->sy + area->height -1;
		val |= 0x00000140;
	}

	tmp = ((area->sx + area->width ) * epinfo.bpp);
	blksrcwidth = tmp / 32;
	if (blksrcwidth > 0 && (tmp % 32 == 0))
		blksrcwidth--;
	blksrcwidth = blksrcwidth - (area->sx * epinfo.bpp) / 32;

	tmp = ((area->dx + area->width) * epinfo.bpp);
	blkdestwidth = tmp / 32;
	if (blkdestwidth > 0 && (tmp % 32 == 0))
		blkdestwidth--;
	blkdestwidth = blkdestwidth - (area->dx * epinfo.bpp) / 32;

	outl(0x00000000, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);

	/*** src ***/
	outl((((startsx * epinfo.bpp) & 0x1F) |
	      (((stopsx * epinfo.bpp ) & 0x1F) << 16)),
	     EP93XX_GRAPHICS_ACCEL_SRCPIXELSTRT);
	outl((epinfo.fb_phys + (startsy * epinfo.xres * epinfo.bpp ) / 8 +
	      (startsx * epinfo.bpp ) / 8),
	     EP93XX_GRAPHICS_ACCEL_BLKSRCSTRT);
	outl(((epinfo.xres * epinfo.bpp) / 32),
	     EP93XX_GRAPHICS_ACCEL_SRCLINELENGTH);
	outl(blksrcwidth, EP93XX_GRAPHICS_ACCEL_BLKSRCWIDTH);

	/*** dest ***/
	outl((((startdx * epinfo.bpp) & 0x1F) |
	      (((stopdx * epinfo.bpp ) & 0x1F) << 16)),
	     EP93XX_GRAPHICS_ACCEL_DESTPIXELSTRT);
	outl((epinfo.fb_phys + (startdy * epinfo.xres * epinfo.bpp ) / 8 +
	      (startdx * epinfo.bpp ) / 8),
	     EP93XX_GRAPHICS_ACCEL_BLKDSTSTRT);
	outl(((epinfo.xres * epinfo.bpp) / 32),
	     EP93XX_GRAPHICS_ACCEL_DESTLINELENGTH);
	outl(blkdestwidth, EP93XX_GRAPHICS_ACCEL_BLKDESTWIDTH);
	outl(area->height - 1, EP93XX_GRAPHICS_ACCEL_BLKDESTHEIGHT);
	outl(epinfo.pixformat | val | 0x00000003, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);
	ep93xxfb_wait();
}


int
ep93xxfb_imageblit(struct fb_info *info, struct ep93xx_blit *blit)
{
	static unsigned char pucBlitBuf[4096];
	unsigned long value = 0;
	unsigned long size;
	unsigned long dx1, dx2, dy1, sx1, sx2, sy1;

	if ((blit->dx >= info->var.xres) ||
	    (blit->dy >= info->var.yres) ||
	    ((blit->dx + blit->width - 1) >= info->var.xres) ||
	    ((blit->dy + blit->height - 1) >= info->var.yres))
		return -EFAULT;

	value = blit->flags & (BLIT_TRANSPARENT | BLIT_MASK_MASK |
			       BLIT_DEST_MASK | BLIT_1BPP_SOURCE);

	if ((blit->flags & BLIT_SOURCE_MASK) == BLIT_SOURCE_MEMORY) {
		if (blit->flags & BLIT_1BPP_SOURCE) {
			size = blit->swidth * blit->height / 8;
			if (size <= 4096) {
				if (copy_from_user(pucBlitBuf, blit->data, size))
					return -EFAULT;
			} else {
				for (size = blit->height; size; ) {
					blit->height = 32768 / blit->swidth;
					if (blit->height > size)
						blit->height = size;
					ep93xxfb_imageblit(info, blit);
					blit->dy += blit->height;
					blit->data += blit->swidth *
						      blit->height / 8;
					size -= blit->height;
				}
				return 0;
			}
			if (blit->flags & BLIT_TRANSPARENT)
				value ^= 0x00004000;
			outl(0x00000007, EP93XX_GRAPHICS_ACCEL_SRCPIXELSTRT);
			outl(blit->swidth >> 5,
			     EP93XX_GRAPHICS_ACCEL_SRCLINELENGTH);
			outl((blit->width - 1) >> 5,
			     EP93XX_GRAPHICS_ACCEL_BLKSRCWIDTH);
			outl(blit->bgcolor, EP93XX_GRAPHICS_ACCEL_BACKGROUND);
		} else {
			size = blit->swidth * blit->height * epinfo.bpp / 8;
			if (size <= 4096) {
				if (copy_from_user(pucBlitBuf, blit->data, size))
					return -EFAULT;
			} else {
				for (size = blit->height; size; ) {
					blit->height = 4096 / epinfo.bpp / 8 /
						blit->swidth;
					if (blit->height > size)
						blit->height = size;
					ep93xxfb_imageblit(info,blit);
					blit->dy += blit->height;
					blit->data += blit->swidth *
						      blit->height *
						      epinfo.bpp / 8;
					size -= blit->height;
				}
				return 0;
			}
			outl(0x00000000, EP93XX_GRAPHICS_ACCEL_SRCPIXELSTRT);
			outl(blit->swidth >> pixel_shift,
			     EP93XX_GRAPHICS_ACCEL_SRCLINELENGTH);
			outl((blit->width - 1) >> pixel_shift,
			     EP93XX_GRAPHICS_ACCEL_BLKSRCWIDTH);
		}
		outl(virt_to_phys(pucBlitBuf), EP93XX_GRAPHICS_ACCEL_BLKSRCSTRT);
		dx1 = blit->dx;
		dy1 = blit->dy;
		dx2 = blit->dx + blit->width - 1;
	} else {	//SOURCE_SCREEN
		if ((blit->sx >= info->var.xres) ||
		    (blit->sy >= info->var.yres) ||
		    ((blit->sx + blit->width - 1) >= info->var.xres) ||
		    ((blit->sy + blit->height - 1) >= info->var.yres))
			return -EFAULT;

		if ((blit->dy == blit->sy) && (blit->dx == blit->sx))
			return 0;

		if ((blit->dy == blit->sy) && (blit->dx > blit->sx) &&
		    (blit->dx < (blit->sx + blit->width - 1))) {
			dx1 = blit->dx + blit->width - 1;
			dx2 = blit->dx;
			sx1 = blit->sx + blit->width - 1;
			sx2 = blit->sx;
			value |= 0x000000a0;
		} else {
			dx1 = blit->dx;
			dx2 = blit->dx + blit->width - 1;
			sx1 = blit->sx;
			sx2 = blit->sx + blit->width - 1;
		}

		if (blit->dy <= blit->sy) {
			dy1 = blit->dy;
			sy1 = blit->sy;
		} else {
			dy1 = blit->dy + blit->height - 1;
			sy1 = blit->sy + blit->height - 1;
			value |= 0x00000140;
		}

		outl(((sx1 & pixel_mask) * epinfo.bpp) |
		     (((sx2 & pixel_mask) * epinfo.bpp) << 16),
		     EP93XX_GRAPHICS_ACCEL_SRCPIXELSTRT);

		outl(epinfo.fb_phys + (sy1 *(info->var.xres * epinfo.bpp / 8)) +
		     ((sx1 * epinfo.bpp / 8) & ~pixel_mask),
		     EP93XX_GRAPHICS_ACCEL_BLKSRCSTRT);

		outl((info->var.xres * epinfo.bpp / 8)/ 4,
		     EP93XX_GRAPHICS_ACCEL_SRCLINELENGTH);
		if (sx1 < sx2)
			outl((sx2 >> pixel_shift) - (sx1 >> pixel_shift),
			     EP93XX_GRAPHICS_ACCEL_BLKSRCWIDTH);
		else
			outl((sx1 >> pixel_shift) - (sx2 >> pixel_shift),
			     EP93XX_GRAPHICS_ACCEL_BLKSRCWIDTH);
	}

	outl(((dx1 & pixel_mask) * epinfo.bpp) |
	     (((dx2 & pixel_mask) * epinfo.bpp) << 16),
	     EP93XX_GRAPHICS_ACCEL_DESTPIXELSTRT);
	outl(epinfo.fb_phys + (dy1 *(info->var.xres * epinfo.bpp / 8)) +
             ((dx1 * epinfo.bpp / 8) & ~pixel_mask), EP93XX_GRAPHICS_ACCEL_BLKDSTSTRT);

	outl((info->var.xres * epinfo.bpp / 8) / 4, EP93XX_GRAPHICS_ACCEL_DESTLINELENGTH);

	if (dx1 < dx2)
		outl((dx2 >> pixel_shift) - (dx1 >> pixel_shift),
		     EP93XX_GRAPHICS_ACCEL_BLKDESTWIDTH);
	else
		outl((dx1 >> pixel_shift) - (dx2 >> pixel_shift),
		     EP93XX_GRAPHICS_ACCEL_BLKDESTWIDTH);

	outl(blit->height - 1, EP93XX_GRAPHICS_ACCEL_BLKDESTHEIGHT);
	outl(blit->fgcolor, EP93XX_GRAPHICS_ACCEL_BLOCKMASK);
	outl(blit->transcolor, EP93XX_GRAPHICS_ACCEL_TRANSPATTRN);
	outl(value | pixel_format | 0x00000003, EP93XX_GRAPHICS_ACCEL_BLOCKCTRL);

	ep93xxfb_wait();

	return 0;
}
